 /*
 * Connections        - Arduino Pro Mini 5V/16MHz    - Arduino Pro Mini 3.V/8MHz
 * -------------------------------------------------------------------------------
 * 
 * Receiver
 * Ext -              - GND                           - GND
 * Ext +              - RAW/Vin                       - RAW/Vin
 * Ext Sig            - TX - R2k4 - RX                - TX - R2k4 - RX
 * 
 * GY-63 / MS5611 (cover sensor with black heat shrink tube as it is very sensitive to sun light)
 * VCC                - VCC/5V                        - VCC/3.3V
 * GND                - GND                           - GND
 * SCL                - A5                            - A5
 * SDA                - A4                            - A4
 * 
 * Voltage divider rx battery 1/2
 * battery -          - GND                           - GND
 * battery +          - R2k4 - A6/A7 - R2k4 - GND     - R4k7 - A6/A7 - R2k4 - GND
 *                      max. battery voltage: 10V       max. battery voltage: 9.76V
 * 
 * AttoPilot Voltage and Current Sense Breakout
 * GND                - GND                           - GND
 * V                  - A0                            - A0
 * I                  - A1                            - A1
 * 
 * ArduPilot Analog Pitot Airspeed Sensor / MPXV7002 (needs 5V Arduino)
 * GND                - GND
 * VS                 - VCC/5
 * VOUT               - A3
 */

//#define ARDUINO_3.3V_8MHZ // uncomment when using a 3.3V/8MHz Arduino
#ifdef ARDUINO_3.3V_8MHZ
  #define VCC 3.3 // V
  #define F_CPU 8000000L // Hz
#else
  #define VCC 5.0 // V
  #define F_CPU 16000000L // Hz
#endif
#define SAMPLING_PERIOD 15 // ms
#define SAMPLES_ALTITUDE 100 // 100: good compromise between reaction time and noise level of vario, for faster reaction decrease number, to reduce noise increase number
#define FILTER_CUTOFF_FREQUENCY 10 // Hz
#define CALIBRATION_OVERSAMPLINGRATE 128
#define DRIVE_CURRENT_MAX_HALF_LIFE 600 // s 

#define RX_VOLTAGE_1_PIN A6
#define RX_VOLTAGE_2_PIN A7 
#define DRIVE_VOLTAGE_PIN A0
#define DRIVE_CURRENT_PIN A1
#define PITOT_PRESSURE_PIN A3

#include "JetiExProtocol.h"
#include <EEPROM.h>

JetiExProtocol jetiEx;

unsigned long timeSample; // us
float airTemperature, airTemperature0; // Kelvin
float airPressure, airPressure0; // Pa
float airDensity; // kg/m^3
float altitude, altitudeBuffer[SAMPLES_ALTITUDE]; // m
float climbRate; // m/s
float rxVoltage1, rxVoltage2, driveVoltage, driveVoltageLow, driveVoltageHigh; // V
float driveCurrent, driveCurrentMax, driveCurrentLow, driveCurrentHigh; // A
const float driveCurrentMaxAttenuation = pow(0.5, SAMPLING_PERIOD / 1000 / DRIVE_CURRENT_MAX_HALF_LIFE);
float drivePower; // W
float driveCapacity; // Ah
float driveInternalResistance; // mOhm
float pitotPressure; // Pa
float airspeedSquared; // m^2/s^2
float airspeed; // m/s
int samplingPeriodMargin; // ms

struct settings {
  bool initEEPROM;
  bool varioTEC;
  float rxVoltage1Scale;
  float rxVoltage2Scale;
  float driveVoltageScale;
  float driveCurrentOffset;
  float driveCurrentScale;
  float pitotPressureOffset;
  float pitotPressureScale;
} set;

JETISENSOR_CONST sensors[] PROGMEM = {
 //id   name              unit      data type               precision 0->0, 1->0.0, 2->0.00
  { 1,  "Rel. Altit",     "m",      JetiSensor::TYPE_14b,   0},
  { 2,  "Climb Rate",     "m/s",    JetiSensor::TYPE_14b,   2},
  { 3,  "RX Voltage 1",   "V",      JetiSensor::TYPE_14b,   2},
  { 4,  "RX Voltage 2",   "V",      JetiSensor::TYPE_14b,   2},
  { 5,  "Drive Voltage",  "V",      JetiSensor::TYPE_14b,   1},
  { 6,  "Drive Current",  "A",      JetiSensor::TYPE_14b,   1},
  { 7,  "Drive Power",    "W",      JetiSensor::TYPE_14b,   0},
  { 8,  "Drive Capacity", "Ah",     JetiSensor::TYPE_14b,   2},
  { 9,  "Drive IR",       "mOhm",   JetiSensor::TYPE_14b,   1},
  {10,  "Airspeed",       "m/s",    JetiSensor::TYPE_14b,   2},
  //{11,  "Debug",          "",       JetiSensor::TYPE_14b,   0},
  0 // end of array
};

void setup() {
  pinMode(LED_BUILTIN, OUTPUT);
  digitalWrite(LED_BUILTIN, HIGH);
  
  EEPROM.get(0, set);
  if (set.initEEPROM) initEEPROM();
  
  jetiEx.Start("Esens", sensors);
  jetiEx.SetJetiboxText(JetiExProtocol::LINE1, "Esens V1.1");
  jetiEx.SetJetiboxText(JetiExProtocol::LINE2, "Welcome");

  MS5611Init();
  initAltitude();
}

void loop() {
  timeSample += SAMPLING_PERIOD;
  samplingPeriodMargin = timeSample - millis();
  digitalWrite(LED_BUILTIN, samplingPeriodMargin <= 0); // switch led on if sampling period is too short
  while (long(timeSample - millis()) > 0); // wait for next sample

  MS5611GetTemperaturePressure(airTemperature, airPressure);
  altitude = airTemperature0 / 0.0065 * (1 - pow(airPressure / airPressure0, 1 / 5.255)); // m, Source: Wikipedia Deutsch - Barometrische Höhenformel - Typische Temperaturgradienten
  airDensity = airPressure / (287.058 * airTemperature); // kg/m^3, Source: Wikipedia Deutsch - Luftdichte - Höhenabhängigkeit

  pitotPressure = set.pitotPressureScale * (analogReadOversampling(PITOT_PRESSURE_PIN, 16) - set.pitotPressureOffset);
  if (pitotPressure > 0) {
    airspeedSquared = 2.0 * pitotPressure / airDensity; // m^2/s^2
    airspeed = sqrt(airspeedSquared); // m/s
  } else {
    airspeedSquared = 0;
    airspeed = 0;
  }
  
  float x = -0.5 * (SAMPLES_ALTITUDE - 1), sxx = 0, sxy = 0; // init linear regression of altitude samples
  for (int i = 0;  i < SAMPLES_ALTITUDE - 1; i++) { // build sums over all altitude samples
    altitudeBuffer[i] = altitudeBuffer[i + 1]; // shift altitude data buffer by one sample
    sxx += x * x;
    sxy += x * altitudeBuffer[i];
    x++;
  }
  altitudeBuffer[SAMPLES_ALTITUDE - 1] = altitude;
  if (set.varioTEC) altitudeBuffer[SAMPLES_ALTITUDE - 1] += airspeedSquared / (2 * 9.81); // altitude which could be reached by pulling up the plane (all kinetic energy is converted into potential energy)
  sxx += x * x;
  sxy += x * altitudeBuffer[SAMPLES_ALTITUDE - 1];
  
  climbRate = sxy * 1000 / (sxx * SAMPLING_PERIOD); // calculate climb rate as slope of linear regression

  filter(rxVoltage1, set.rxVoltage1Scale * analogRead(RX_VOLTAGE_1_PIN)); // V
  filter(rxVoltage2, set.rxVoltage2Scale * analogRead(RX_VOLTAGE_2_PIN)); // V
  filter(driveVoltage, set.driveVoltageScale * analogRead(DRIVE_VOLTAGE_PIN)); // V  
  filter(driveCurrent, set.driveCurrentScale * (float(analogRead(DRIVE_CURRENT_PIN)) - set.driveCurrentOffset)); // A
  drivePower = driveVoltage * driveCurrent; // W
  driveCapacity += driveCurrent * SAMPLING_PERIOD / 3600000; // Ah

  if (driveCurrentMax < driveCurrent) {
    driveCurrentMax = driveCurrent;
  } else
    driveCurrentMax *= driveCurrentMaxAttenuation;
    
  if (driveCurrent < 0.05 * driveCurrentMax) {
    driveCurrentLow = driveCurrent;
    driveVoltageHigh = driveVoltage;
  } else if (driveCurrent > 0.95 * driveCurrentMax) {
    driveCurrentHigh = driveCurrent;
    driveVoltageLow = driveVoltage;
  }
  driveInternalResistance = 1000 * (driveVoltageHigh - driveVoltageLow) / (driveCurrentHigh - driveCurrentLow); // mOhm
  
  handleJetiBox();
  
  jetiEx.SetSensorValue(1, altitude);
  jetiEx.SetSensorValue(2, climbRate * 100);
  jetiEx.SetSensorValue(3, rxVoltage1 * 100);
  jetiEx.SetSensorValue(4, rxVoltage2 * 100);
  jetiEx.SetSensorValue(5, driveVoltage * 10);
  jetiEx.SetSensorValue(6, driveCurrent * 10);
  jetiEx.SetSensorValue(7, drivePower);
  jetiEx.SetSensorValue(8, driveCapacity * 100);
  jetiEx.SetSensorValue(9, driveInternalResistance * 10);
  jetiEx.SetSensorValue(10, airspeed * 100);
  //jetiEx.SetSensorValue(11 , samplingPeriodMargin);

  jetiEx.DoJetiSend();
}

float analogReadOversampling(int pin, int samples) {
  unsigned long sum = 0;
  for (int i = 0;  i < samples; i++) sum += long(analogRead(pin));
  return float(sum) / samples;  
}

void filter(float &filteredValue, float newValue) {
  static float alpha = 2 * PI * 0.001 * SAMPLING_PERIOD * FILTER_CUTOFF_FREQUENCY / (2 * PI * 0.001 * SAMPLING_PERIOD * FILTER_CUTOFF_FREQUENCY + 1);
  static float beta = 1 - alpha;
  filteredValue = alpha * newValue + beta * filteredValue;
}

void initAltitude() {
  airTemperature0 = 0;
  airPressure0 = 0;
  for (int i = 0;  i < 64; i++) {
    MS5611GetTemperaturePressure(airTemperature, airPressure);
    airTemperature0 += airTemperature;
    airPressure0 += airPressure;
  }
  airTemperature0 /= 64;
  airPressure0 /= 64;
  timeSample = millis();
}

void initEEPROM() {
  set.initEEPROM = false;
  initVarioTEC();
  initRxVoltage1Scale();
  initRxVoltage2Scale();
  initDriveVoltageScale();
  initDriveCurrentOffset();
  initDriveCurrentScale();
  initPitotPressureOffset();
  initPitotPressureScale();
  EEPROM.put(0, set);
}
void initVarioTEC() {
  set.varioTEC = false;
}
void initRxVoltage1Scale() {
  set.rxVoltage1Scale = 10.0 / 1023;
}
void initRxVoltage2Scale() {
  set.rxVoltage2Scale = 10.0 / 1023;
}
void initDriveVoltageScale() {
  set.driveVoltageScale = 15.7 * VCC / 1023; // AttoPilot Voltage and Current Sense Breakout - 90A
}
void initDriveCurrentOffset() {
  set.driveCurrentOffset = 0; // AttoPilot Voltage and Current Sense Breakout - 90A
}
void initDriveCurrentScale() {
  set.driveCurrentScale = 89.4 / 3.3 * VCC / 1023; // AttoPilot Voltage and Current Sense Breakout - 90A
}
void initPitotPressureOffset() {
  set.pitotPressureOffset = 2.5 / VCC * 1023; // MPXV7002
}
void initPitotPressureScale() {
  set.pitotPressureScale = 4000 / 4.0 * VCC / 1023; // MPXV7002
}

